from __future__ import annotations

from parcelforce_expresslink.address import AddressBase, AddressRecipient, Contact as ContactPF
from parcelforce_expresslink.client import ParcelforceClient
from parcelforce_expresslink.combadge import CreateShipmentService
from parcelforce_expresslink.request_response import ShipmentRequest, ShipmentResponse
from parcelforce_expresslink.shipment import Shipment as ShipmentPF
from shipaw.fapi.responses import ShipmentBookingResponse
from shipaw.models.address import Address as AddressAgnost, Contact as ContactAgnost, FullContact

from shipaw.models.services import Services
from shipaw.models.shipment import Shipment as ShipmentAgnost

PARCELFORCE_SERVICES = Services(
    NEXT_DAY='SND',
    NEXT_DAY_12='S12',
    NEXT_DAY_9='09',
)


def address_from_agnostic[addr_type: AddressBase](
    address: AddressAgnost, cls: type[addr_type] = AddressRecipient
) -> addr_type:
    return cls(
        address_line1=address.address_lines[0],
        address_line2=address.address_lines[1] if len(address.address_lines) > 1 else None,
        address_line3=address.address_lines[2] if len(address.address_lines) > 2 else None,
        town=address.town,
        postcode=address.postcode,
        country=address.country,
    )


def address_from_agnostic_fc[addr_type: AddressBase](cls: type[addr_type], full_contact: FullContact) -> addr_type:
    return address_from_agnostic(full_contact.address, cls)


def contact_from_agnostic_fc[contact_type: ContactPF](
    cls: type[contact_type], full_contact: FullContact
) -> contact_type:
    return cls(
        business_name=full_contact.address.business_name,
        contact_name=full_contact.contact.contact_name,
        email_address=full_contact.contact.email_address,
        mobile_phone=full_contact.contact.mobile_phone,
    )


def full_contact_from_provider_contact_address(contact: ContactPF, address: AddressBase) -> FullContact:
    return FullContact(
        address=AddressAgnost(
            business_name=contact.business_name,
            address_lines=[
                line for line in [address.address_line1, address.address_line2, address.address_line3] if line
            ],
            town=address.town,
            postcode=address.postcode,
            country=address.country,
        ),
        contact=ContactAgnost(
            contact_name=contact.contact_name,
            email_address=contact.email_address,
            mobile_phone=contact.mobile_phone,
        ),
    )


def split_string_into_chunks(s, chunk_size):
    return [s[i : i + chunk_size] for i in range(0, len(s), chunk_size)]


def ref_dict_from_str(ref_string: str) -> dict[str, str]:
    refs = split_string_into_chunks(ref_string, 24)
    if len(refs) > 5:
        raise ValueError('Reference too long, maximum 120 characters allowed')
    ref_nums = {f'reference_number{i}': ref for i, ref in enumerate(refs, start=1)}
    return ref_nums


def join_refs(refs: dict[str, str]) -> str:
    refs = [refs.get(f'reference_number{i+1}', '') for i in range(len(refs))]
    return ''.join(refs).strip()


def parcelforce_shipment_from_agnostic(
    shipment: ShipmentAgnost | dict,
) -> ShipmentPF | dict:
    shipment = ShipmentAgnost.model_validate(shipment)
    ship_pf = ShipmentPF(
        **ref_dict_from_str(shipment.reference),
        recipient_contact=contact_from_agnostic_fc(ContactPF, shipment.recipient),
        recipient_address=address_from_agnostic_fc(AddressRecipient, shipment.recipient),
        total_number_of_parcels=shipment.boxes,
        shipping_date=shipment.shipping_date,
        service_code=(PARCELFORCE_SERVICES.lookup(shipment.service)),
    )
    ship_pf = ship_pf.convert(shipment.direction)

    return ship_pf


def parcelforce_shipment_to_agnostic(shipment: ShipmentPF) -> ShipmentAgnost:
    return ShipmentAgnost(
        recipient=full_contact_from_provider_contact_address(shipment.recipient_contact, shipment.recipient_address),
        sender=full_contact_from_provider_contact_address(shipment.sender_contact, shipment.sender_address)
        if shipment.sender_contact and shipment.sender_address
        else None,
        boxes=shipment.total_number_of_parcels,
        shipping_date=shipment.shipping_date,
        direction=shipment.direction,
        reference=', '.join(
            filter(
                None,
                [
                    shipment.reference_number1,
                    shipment.reference_number2,
                    shipment.reference_number3,
                    shipment.reference_number4,
                    shipment.reference_number5,
                ],
            )
        ),
    )


def book_shipment(shipment: dict | ShipmentAgnost) -> ShipmentBookingResponse:
    shipment = ShipmentAgnost.model_validate(shipment)
    shipment_pf = parcelforce_shipment_from_agnostic(shipment)
    shipment_request_pf = ShipmentRequest(requested_shipment=shipment_pf)

    el_client = ParcelforceClient()
    authorized_shipment = shipment_request_pf.authenticate_from_settings()
    ship_req = authorized_shipment.model_dump(by_alias=True)

    back = el_client.backend(CreateShipmentService)
    pf_response: ShipmentResponse = back.createshipment(request=ship_req)
    pf_response.handle_errors()

    resp_agnost = ShipmentBookingResponse(
        shipment=shipment,
        shipment_num=pf_response.shipment_num,
        tracking_link=pf_response.tracking_link(),
        data=pf_response.model_dump(),
        status=pf_response.status,
        success=pf_response.success,
        label_data=el_client.get_label_content(pf_response.shipment_num),
    )

    return resp_agnost



import json
from base64 import b64decode

import httpx
from pydantic import BaseModel

from shipaw.fapi.responses import ShipmentBookingResponse
from shipaw.models.shipment import Shipment, Shipment as ShipmentAgnost
from apc_hypaship.config import apc_settings
from apc_hypaship.models.request.shipment import Shipment as ShipmentAPC
from shipaw.models.provider import ShippingProvider, register_provider
from shipaw.providers.apc.provider_funcs import APC_SERVICES, \
    apc_shipment_from_agnostic, \
    apc_shipment_to_agnostic, \
    shipment_booking_errored


# @dataclass
@register_provider
class APCShippingProvider(ShippingProvider):
    name = 'APC'
    services = APC_SERVICES

    def agnostic_shipment(self, shipment: ShipmentAPC) -> Shipment:
        return apc_shipment_to_agnostic(shipment)

    def provider_shipment(self, shipment: Shipment) -> BaseModel:
        return apc_shipment_from_agnostic(shipment)

    def book_shipment(self, shipment: dict | ShipmentAgnost, settings=apc_settings()) -> ShipmentBookingResponse:
        """Takes provider ShipmnentDict, or ShipmentAgnost object"""
        apc_shipment = self.provider_shipment(shipment)
        shipment_dict = apc_shipment.model_dump(mode='json', by_alias=True)

        res = httpx.post(settings.orders_endpoint, headers=settings.headers, json=shipment_dict)
        res.raise_for_status()
        res_json = res.json()
        messages = json.loads(res.text).get('Orders').get('Order').get('Messages')
        if 'ErrorFields' in messages.keys():
            return shipment_booking_errored(messages, res, res_json, shipment)
        order = res_json['Orders']['Order']
        order_number = order['OrderNumber']

        return ShipmentBookingResponse(
            shipment=shipment,
            shipment_num=order_number,
            tracking_link='NOT IMPLEMENTED',
            data=res_json,
            status=str(res.status_code),
            success=res.is_success,
            label_data=self.get_label_content(order_number),
        )

    def get_label_content(self, shipment_num: str) -> bytes:
        params = {'labelformat': 'PDF'}
        settings = apc_settings()
        label = httpx.get(settings.one_order_endpoint(shipment_num), headers=settings.headers, params=params)
        label.raise_for_status()
        content = label.json()['Orders']['Order']['Label']['Content']
        return b64decode(content)




import os
from functools import wraps

from fastapi import APIRouter, Body
from fastapi.params import Depends, Query
from starlette.requests import Request
from starlette.responses import HTMLResponse

from shipaw.config import shipaw_settings
from shipaw.fapi.alerts import Alerts
from shipaw.fapi.responses import ShipawTemplateResponse
from shipaw.fapi.form_data import shipment_request_form, shipment_request_form_json
from shipaw.fapi.requests import ShipmentRequest
from shipaw.models.shipment import Shipment, sample_shipment
from shipaw.fapi.routes_api import (
    order_results as order_confirm_json,
    order_summary as order_review_json,
    ship_form as ship_form_json,
)

router = APIRouter()


def render_template_response(request: Request, resp: ShipawTemplateResponse) -> HTMLResponse:
    context = resp.template.context
    context['alerts'] = context.get('alerts', Alerts.empty()) + resp.alerts
    if resp.alerts.errors:
        return shipaw_settings().templates.TemplateResponse(
            request=request,
            name='alerts.html',
            context=context,
        )

    return shipaw_settings().templates.TemplateResponse(
        request=request,
        name=resp.template.template_path,
        context=context,
    )


def html_from_json(json_endpoint):
    @wraps(json_endpoint)
    async def wrapper(request: Request, *args, **kwargs):
        resp = await json_endpoint(request, *args, **kwargs)

        return shipaw_settings().templates.TemplateResponse(
            request=request, name=resp.template_path, context=resp.context
        )

    return wrapper


@router.post('/shipping_form', response_class=HTMLResponse)
async def shipping_form(request: Request, shipment: Shipment = Body(...)) -> HTMLResponse:
    res = await ship_form_json(request, shipment)
    return render_template_response(request, res)


@router.post('/order_summary', response_class=HTMLResponse)
async def order_summary(
    request: Request,
    shipment_request: ShipmentRequest = Depends(shipment_request_form),
) -> HTMLResponse:
    res = await order_review_json(request, shipment_request)
    return render_template_response(request, res)


@router.post('/order_results', response_class=HTMLResponse)
async def order_results(
    request: Request,
    shipment_request: ShipmentRequest = Depends(shipment_request_form_json),
) -> HTMLResponse:
    template_response = await order_confirm_json(request, shipment_request)
    return render_template_response(request, template_response)


@router.get('/test', response_class=HTMLResponse)
async def ship(request: Request) -> HTMLResponse:
    shipment = sample_shipment()
    res = await ship_form_json(request, shipment)
    return render_template_response(request, res)


@router.get('/home_mobile_phone', response_class=HTMLResponse)
async def home_mobile_phone():
    mobile_phone = shipaw_settings().mobile_phone
    return f"""
    <input type="tel" id="mobile_phone" name="mobile_phone" value="{mobile_phone}" required>
    """


@router.get('/open-file/{filepath}', response_class=HTMLResponse)
async def open_file(filepath: str):
    os.startfile(filepath)
    return HTMLResponse(content=f'<span>Re</span>')


@router.post('/print-file/{filepath}', response_class=HTMLResponse)
async def print_file(filepath: str):
    os.startfile(filepath, 'print')
    return HTMLResponse(content=f'<span>Re</span>')



from __future__ import annotations

from pydantic import BaseModel

from shipaw.models.provider import ShippingProvider, register_provider
from shipaw.fapi.responses import ShipmentBookingResponse
from shipaw.models.shipment import Shipment, Shipment as ShipmentAgnost

#
from parcelforce_expresslink.client import ParcelforceClient
from parcelforce_expresslink.shipment import Shipment as ShipmentPF
from shipaw.providers.parcelforce.provider_funcs import (
    book_shipment,
    parcelforce_shipment_from_agnostic,
    parcelforce_shipment_to_agnostic,
)


# @dataclass
@register_provider
class ParcelforceShippingProvider(ShippingProvider):
    name = 'PARCELFORCE'

    def provider_shipment(self, shipment: Shipment) -> BaseModel:
        return parcelforce_shipment_from_agnostic(shipment)

    def agnostic_shipment(self, shipment: ShipmentPF) -> Shipment:
        return parcelforce_shipment_to_agnostic(shipment)

    def book_shipment(self, shipment: dict | ShipmentAgnost) -> ShipmentBookingResponse:
        return book_shipment(shipment)

    def get_label_content(self, shipment_num: str) -> bytes:
        el_client = ParcelforceClient()
        return el_client.get_label_content(shipment_num)




from abc import ABC, abstractmethod
from typing import Callable, ClassVar, Self, TYPE_CHECKING

from loguru import logger
from pydantic import BaseModel

# from shipaw.fapi.backend import try_get_write_label
from shipaw.models.logging import log_obj
from shipaw.models.services import Services
from shipaw.models.shipment import Shipment

if TYPE_CHECKING:
    from shipaw.fapi.requests import ShipmentRequest
    from shipaw.fapi.responses import ShipmentBookingResponse


class ConvertableShipment(ABC, BaseModel):
    @classmethod
    @abstractmethod
    def from_agnostic(cls, shipment: Shipment) -> Self:
        raise NotImplementedError

    @abstractmethod
    def to_agnostic(self) -> Shipment:
        raise NotImplementedError


# class ProviderShipment(ConvertableShipment, ABC):
#     agnostic_shipment: Shipment
#     provider_shipment: ConvertableShipment


ProviderShipmentFn = Callable[[Shipment], BaseModel]
AgnosticShipmentFn = Callable[[BaseModel], Shipment]
BookingFn = Callable[[dict | Shipment], 'ShipmentBookingResponseAgnost']

#
# class ShipProv:
#     name: str
#     services: Services
#     provider_shipment: ProviderShipmentFn
#     agnostic_shipment: AgnosticShipmentFn
#     book_shipment: BookingFn
#     get_label_content: Callable[[str], bytes]
#
#     def __init__(
#         self,
#         name: str,
#         services: Services,
#         provider_shipment: ProviderShipmentFn,
#         agnostic_shipment: AgnosticShipmentFn,
#         book_shipment: BookingFn,
#         get_label_content: Callable[[str], bytes],
#     ) -> None:
#         self.name = name
#         self.services = services
#         self.provider_shipment = provider_shipment
#         self.agnostic_shipment = agnostic_shipment
#         self.book_shipment = book_shipment
#         self.get_label_content = get_label_content


# @dataclass
class ShippingProvider(ABC):
    name: ClassVar[str]
    # service_map: ClassVar[MappingProxyType]
    services: Services

    @staticmethod
    @abstractmethod
    def provider_shipment(shipment: Shipment) -> BaseModel:
        """Takes agnostic Shipment object and returns provider Shipment object"""
        ...

    @staticmethod
    @abstractmethod
    def agnostic_shipment(shipment: BaseModel) -> Shipment:
        """Takes provider Shipment object and returns agnostic Shipment object"""
        ...

    @staticmethod
    @abstractmethod
    def book_shipment(shipment: dict | Shipment) -> 'ShipmentBookingResponse': ...

    @staticmethod
    @abstractmethod
    def get_label_content(shipment_num: str) -> bytes: ...

    @staticmethod
    def handle_response(request: 'ShipmentRequest', response: 'ShipmentBookingResponse'):
        # log_booked_shipment(request, response)
        log_obj(response)

    @staticmethod
    async def handle_response_async(request: 'ShipmentRequest', response: 'ShipmentBookingResponse'):
        log_obj(response, 'Shipment Booked')
        try:
            if response.label_data is None and response.shipment_num:
                response.label_data = request.provider.get_label_content(response.shipment_num)
            await response.write_label_file()
        except Exception as e:
            logger.exception(f'Error getting or writing label data: {e}')

    def get_shipment_alias_dict(self, shipment: Shipment | dict) -> dict:
        shipment = shipment.model_validate(shipment)
        shipment = self.provider_shipment(shipment)
        shipment_alias = shipment.model_dump(mode='json', by_alias=True)
        return shipment_alias


PROVIDER_REGISTER: dict[str, type[ShippingProvider]] = {}


def register_provider(cls: type[ShippingProvider]) -> type[ShippingProvider]:
    PROVIDER_REGISTER[str(cls.name)] = cls
    return cls



from shipaw.providers.apc.provider import APCShippingProvider
from shipaw.providers.parcelforce.provider import ParcelforceShippingProvider

__all__ = ['APCShippingProvider', 'ParcelforceShippingProvider']


from dataclasses import dataclass
from typing import Literal


@dataclass
class Services:
    NEXT_DAY: str
    NEXT_DAY_12: str
    NEXT_DAY_9: str

    def reverse_lookup(self, service_code: str) -> str | None:
        res = next((name for name, code in self.__dict__.items() if code == service_code), None)
        if not res:
            raise ValueError(f'Invalid service code: {service_code}')
        return res

    def lookup(self, service_name: str) -> str:
        res = self.__dict__.get(service_name)
        if not res:
            raise ValueError(f'Invalid service name: {service_name}')
        return res


ServiceType = Literal['NEXT_DAY', 'NEXT_DAY_12', 'NEXT_DAY_9', 'SATURDAY']

SERVICE_ATTRS = tuple(Services.__annotations__.keys())
# type ServiceDict = dict[ServiceType, str]


# def get_agnostic_service_name(service_code: str, services: Services) -> ServiceType | None:
#     service_map: ServiceDict = {
#         'NEXT_DAY': services.NEXT_DAY,
#         'NEXT_DAY_12': services.NEXT_DAY_12,
#         'NEXT_DAY_9': services.NEXT_DAY_9,
#         'SATURDAY': 'Saturday Delivery',
#     }
#     for service_name, code in service_map.items():
#         if code == service_code:
#             return service_name
#     return None


from apc_hypaship.models.request.address import Address, Contact
from apc_hypaship.models.request.shipment import GoodsInfo, Order, Orders, Shipment as ShipmentAPC, ShipmentDetails
from shipaw.fapi.alerts import Alert, Alerts
from shipaw.fapi.responses import ShipmentBookingResponse
from shipaw.models.address import Address as AddressAgnost, Contact as ContactAgnost, FullContact
from shipaw.models.services import Services
from shipaw.models.ship_types import ShipDirection
from shipaw.models.shipment import Shipment as ShipmentAgnost

APC_SERVICES = Services(
    NEXT_DAY='ND16',
    NEXT_DAY_12='ND12',
    NEXT_DAY_9='ND09',
)


def address_from_agnostic_fc[addr_type: Address](cls: type[addr_type], full_contact: FullContact) -> addr_type:
    lines_ = [_ for _ in full_contact.address.address_lines[1:] if _]
    lines = ', '.join(lines_)
    return cls(
        company_name=full_contact.address.business_name,
        address_line_1=full_contact.address.address_lines[0],
        address_line_2=lines,
        city=full_contact.address.town,
        postal_code=full_contact.address.postcode,
        country_code=full_contact.address.country,
        contact=Contact(
            person_name=full_contact.contact.contact_name,
            email=full_contact.contact.email_address,
            mobile_number=full_contact.contact.mobile_phone,
            phone_number=full_contact.contact.phone_number or full_contact.contact.mobile_phone,
        ),
    )


def contact_from_agnostic_fc[contact_type: Contact](cls: type[contact_type], full_contact: FullContact) -> contact_type:
    return cls(
        person_name=full_contact.contact.contact_name,
        email=full_contact.contact.email_address,
        mobile_number=full_contact.contact.mobile_phone,
        phone_number=full_contact.contact.phone_number,
    )


def full_contact_from_apc_contact_address(contact: Contact, address: Address) -> FullContact:
    return FullContact(
        address=AddressAgnost(
            business_name=address.company_name,
            address_lines=[line for line in [address.address_line_1, address.address_line_2] if line],
            town=address.city,
            postcode=address.postal_code,
            country=address.country_code,
        ),
        contact=ContactAgnost(
            contact_name=contact.person_name,
            email_address=contact.email,
            mobile_phone=contact.mobile_number,
            phone_number=contact.phone_number or contact.mobile_number,
        ),
    )


def apc_shipment_to_agnostic(shipment: ShipmentAPC) -> ShipmentAgnost:
    order = shipment.orders.order
    del_fc = full_contact_from_apc_contact_address(order.delivery.contact, order.delivery)
    send_fc = (
        full_contact_from_apc_contact_address(order.collection.contact, order.collection) if order.collection else None
    )
    service = APC_SERVICES.reverse_lookup(order.product_code)

    return ShipmentAgnost(
        service=service,
        shipping_date=order.collection_date,
        reference=order.reference,
        recipient=del_fc,
        sender=send_fc,
        boxes=order.shipment_details.number_of_pieces,
        direction=ShipDirection.INBOUND if order.collection is not None else ShipDirection.OUTBOUND,
        collect_ready=order.ready_at,
        collect_closed=order.closed_at,
    )


def apc_shipment_from_agnostic(shipment: ShipmentAgnost) -> ShipmentAPC:
    # todo handle direction
    if shipment.direction not in [ShipDirection.INBOUND, ShipDirection.OUTBOUND]:
        raise NotImplementedError('APCShippingProvider does not support DROPOFF shipments')

    service_code = APC_SERVICES.lookup(shipment.service)
    ship_deets = ShipmentDetails(number_of_pieces=shipment.boxes)

    order = Order(
        ready_at=shipment.collect_ready,
        closed_at=shipment.collect_closed,
        collection_date=shipment.shipping_date,
        product_code=service_code,
        reference=shipment.reference,
        delivery=address_from_agnostic_fc(Address, shipment.recipient),
        collection=address_from_agnostic_fc(Address, shipment.sender) if shipment.sender else None,
        goods_info=GoodsInfo(),
        shipment_details=ship_deets,
    )
    return ShipmentAPC(orders=Orders(order=order))


def shipment_booking_errored(messages, res, res_json, shipment):
    fieldname = messages['ErrorFields']['ErrorField']['FieldName']
    message = messages['ErrorFields']['ErrorField']['ErrorMessage']
    alerts = Alerts(alert=[Alert(message=f'Error booking shipment: {fieldname}: {message}')])
    return ShipmentBookingResponse(
        alerts=alerts,
        shipment=shipment,
        shipment_num='FAILED TO BOOK',
        tracking_link='NOT IMPLEMENTED',
        data=res_json,
        status=str(res.status_code),
        success=False,
        label_data=b'',
    )



import json

from httpx import HTTPStatusError
from loguru import logger
from parcelforce_expresslink.client import ParcelforceClient

from shipaw.fapi.requests import ShipmentRequest
from shipaw.fapi.responses import ShipmentBookingResponse
from shipaw.fapi.alerts import AlertType, Alert


def extract_http_error_message(exception: HTTPStatusError) -> str:
    if hasattr(exception, 'response') and exception.response is not None:
        return exception.response.text
    logger.warning('HTTPStatusError has no response attribute')
    return str(exception)


def extract_http_error_message_json(exception: HTTPStatusError) -> dict:
    error_string = extract_http_error_message(exception)
    try:
        error_data = json.loads(error_string)
        return error_data.get('Messages')

    except json.JSONDecodeError:
        logger.warning('Error.response.text is not valid JSON')
        return {'Code': error_string, 'Description': ''}


async def http_status_alerts(exception: HTTPStatusError) -> list[Alert]:
    error_dict = extract_http_error_message_json(exception)
    return [Alert(message=f'{error_dict.get('Code')}:  {error_dict.get('Description')}', type=AlertType.ERROR)]


async def try_book_shipment(shipment_request: ShipmentRequest) -> ShipmentBookingResponse:
    shipment_response = ShipmentBookingResponse(shipment=shipment_request.shipment)
    try:
        shipment_response = shipment_request.provider.book_shipment(shipment_request.shipment)

    except HTTPStatusError as e:
        for alert in await http_status_alerts(e):
            shipment_response.alerts += alert

    except Exception as e:
        logger.exception(f'Error booking shipment: {e}')
        shipment_response.alerts += Alert.from_exception(e)

    return shipment_response


async def try_get_label_data(request: ShipmentRequest, response: ShipmentBookingResponse) -> bytes | None:
    try:
        if response.label_data is None and response.shipment_num:
            response.label_data = request.provider.get_label_content(response.shipment_num)
        return response.label_data

    except HTTPStatusError as e:
        for alert in await http_status_alerts(e):
            response.alerts += alert
        logger.exception(f'HTTP error getting label data')

    except Exception as e:
        logger.exception(f'Error getting label data')
        response.alerts += Alert.from_exception(e)

    return None


async def try_write_label(response: ShipmentBookingResponse):
    try:
        await response.write_label_file()
    except Exception as e:
        logger.exception(f'Error writing label file: {e}')
        response.alerts += Alert.from_exception(e)


async def try_get_write_label(request: ShipmentRequest, response: ShipmentBookingResponse):
    if not response.label_data:
        await try_get_label_data(request, response)
    await try_write_label(response)


def get_el_client() -> ParcelforceClient:
    try:
        return ParcelforceClient()
    except Exception as e:
        logger.error(f'Error getting Parcelforce ExpressLink Client: {e}')
        raise




from __future__ import annotations

import re
from enum import StrEnum
import datetime as dt
from typing import Annotated

import phonenumbers
from loguru import logger
from pydantic import ValidationError, AfterValidator, BeforeValidator, Field


class ShipDirection(StrEnum):
    INBOUND = 'in'
    OUTBOUND = 'out'
    DROPOFF = 'dropoff'


TOD = dt.date.today()
COLLECTION_CUTOFF = dt.time(23, 59, 59)
ADVANCE_BOOKING_DAYS = 28
WEEKDAYS_IN_RANGE = [
    TOD + dt.timedelta(days=i) for i in range(ADVANCE_BOOKING_DAYS) if (TOD + dt.timedelta(days=i)).weekday() < 5
]

COLLECTION_WEEKDAYS = [i for i in WEEKDAYS_IN_RANGE if not i == TOD]


def limit_daterange_no_weekends(v: dt.date) -> dt.date:
    logger.debug(f'Validating date: {v}')
    if v:
        if isinstance(v, str):
            logger.debug(f'parsing date string assuming isoformat: {v}')
            v = dt.date.fromisoformat(v)

        if isinstance(v, dt.date):
            if v < TOD or v.weekday() > 4:
                logger.debug(f'Date {v} is a weekend or in the past - using next weekday')
                v = min(WEEKDAYS_IN_RANGE)

            if v > max(WEEKDAYS_IN_RANGE):
                logger.debug(f'Date {v} is too far in the future - using latest weekday (max 28 days in advance)')
                v = max(WEEKDAYS_IN_RANGE)

    return v


def validate_phone(v: str, values) -> str:
    logger.warning(f'Validating phone: {v}')
    phone = v.replace(' ', '')
    nummy = phonenumbers.parse(phone, 'GB')
    assert phonenumbers.is_valid_number(nummy)
    return phonenumbers.format_number(nummy, phonenumbers.PhoneNumberFormat.E164)


POSTCODE_PATTERN = re.compile(r'([A-Z]{1,2}\d[A-Z\d]? ?\d[A-Z]{2})')

pc_excluded = {'C', 'I', 'K', 'M', 'O', 'V'}


def validate_uk_postcode(v: str):
    if not re.match(POSTCODE_PATTERN, v) and not set(v[-2:]).intersection(pc_excluded):
        raise ValidationError('Invalid UK postcode')
    return v


VALID_POSTCODE = Annotated[
    str,
    AfterValidator(validate_uk_postcode),
    BeforeValidator(lambda s: s.strip().upper()),
    Field(..., description='A valid UK postcode'),
]
